# encoding: utf-8

'''
LDA wrapper script that uses gensim package http://radimrehurek.com/gensim/index.html
Written by Francisco C. Pereira, August 2013
 

'''


import logging, gensim, re
from gensim import corpora, models, similarities, parsing

logging.basicConfig(format='%(asctime)s : %(levelname)s : %(message)s', level=logging.INFO)




filename="events_all"
NUM_TOPICS = 25

replacements=[['\xe9', 'e'],["\?s", "'s"], ["&amp;","and"]]

DEBUG_LDA=False

         


# load events
# Expects a csv file with 4 fields: event Id, venue, categories (list), title, description 
def load_events(filename):
    f = file(filename+".csv")
    f.readline() # headers
    headers= ["date","event_id","source","venue","title","stdate","endate","repeated_count","lat","lng","addr","categories","tags","price","description"]
    #headers = ["event_id", "venue", "categories", "title", "description"]
    events = []
    
    for l in f:
        fields = l.strip().split(",")
        event = dict([ (headers[i], fields[i].strip('"')) for i in range(len(headers)) ])
        #event["start"] = datetime.datetime.strptime( event["start"], "%Y-%m-%d %H:%M:%S" )
        #event["end"] = datetime.datetime.strptime( event["end"], "%Y-%m-%d %H:%M:%S" )
        events.append(event)
    f.close()
    print len(events), "events loaded"
    return events

#Usual text processing, this one is simple. Just stemming and then some replacements of simple encoding bugs
def preprocess(text, stoplist):
    #print "preprocessing ", text
  
    
    text=parsing.stem(text)
    
    
    for r in replacements:   #replace same patterns AFTER cleaning
        p=re.compile(r[0])
        text=p.sub(r[1],text)
           
    return text


# load stoplist
# Stopword list, this can be very important!
def load_stoplist():
    stoplist=[]
    for l in file("stopwords.txt"):
        if "|" in l:
            l = l[:l.index("|")]
        l = l.strip().lower()
        stoplist.append(l)
    return stoplist
    
    


# convert to bag-of-words, remove non-words
def to_bow(events, stoplist):
    words_count = {}

    documents = []
    for e in events:
        doc = []
        words = set([ x.lower() for x in e["description"].split() ])
        for w in words:
            w = w.lower()
            w = w.strip(".;!?\'()+-:*")
            if len(re.sub("[^\w]", "", w)) == 0 or len(w) <= 1: # remove words who are just symbols
                continue
            
            
            
            if not w in stoplist:
                w=preprocess(w, stoplist)
                doc.append(w)
                
                # word frequency
                if not w in words_count:
                    words_count[w] = 1
                else:
                    words_count[w] += 1
        documents.append(doc)
    
    # remove words that appear only once
    for i in range(len(documents)):
        new_doc = []
        for w in documents[i]:
            if words_count[w] > 1:
                new_doc.append(w)
        documents[i] = new_doc
        
    print "once-ocurred words removed"
    print documents
    dictionary = corpora.Dictionary(documents)
    dictionary.save('events_dictionary.dict') # store the dictionary, for future reference
    #dictionary = corpora.Dictionary.load("events_dictionary.dict")
    #print "dictionary created and saved."

    mm = [dictionary.doc2bow(text) for text in documents ]
    corpora.MmCorpus.serialize('documents.mm', mm) # store to disk, for later use
    #mm = gensim.corpora.MmCorpus('documents.mm')
    #print "corpus created and saved."
    return mm, dictionary



        

def generate_LDA_model(events, stoplist, filename, passes=20):
    
    # extract NUM_TOPICS LDA topics, using 20 pass (default) and updating once every 1 chunk
    mm, dictionary=to_bow(events, stoplist)
    
    lda = gensim.models.ldamodel.LdaModel(corpus=mm, id2word=dictionary, num_topics=NUM_TOPICS, update_every=1, chunksize=1000, passes=passes)
    
    lda.save('lda_%d.model' % NUM_TOPICS)
    
    print "LDA Model saved."
    print "done."
    
#Another data preparation step
def prepare_for_LDA_assignment(message, stoplist):
    doc = []
    words = set([ x.lower() for x in message.split() ])
    for w in words:
        w = w.lower()
        w = w.strip(".;!?\'()+-:*")
        if len(re.sub("[^\w]", "", w)) == 0 or len(w) <= 1: # remove words who are just symbols
            continue
        
        if not w in stoplist:
            w=preprocess(w, stoplist)
            doc.append(w)

    return doc

def list2string(words):
    text=""
    for w in words:
        text+=w+" "
        
    return text

#NOTE: this code is somewhat redundant (e.g. it re-loads the LDA model) because 
# it allows you to just assign topics without having to retrain LDA all the time
# For example, you can train the model once, and then comment line "generate_LDA_model(events, stoplist, filename, 1000)" on the 
# main code below  
def generate_topic_assignment(events, stoplist, filename, model):
    lda = gensim.models.ldamodel.LdaModel.load(model)
    
    #Just for reporting
    lda.print_topics(25)
    
    dictionary = corpora.Dictionary.load("events_dictionary.dict")
    #print dictionary
    
    fl=open(filename, "w")
    
    #ID, venue, topics1-25 
    fl.write("event_id, venue, ")
    
    for tid in range(NUM_TOPICS):
        fl.write("topic"+str(tid+1)+",")
    fl.write("category\n")
    
    
    
    for event in events:
        if DEBUG_LDA:
            print "Assigning topics to incident ",event["event_id"]
     
        fl.write(str(event['event_id']))    
        fl.write(","+str(event['venue']))
        
        
        mtext=prepare_for_LDA_assignment(event['description'], stoplist)
        bow=dictionary.doc2bow(mtext)    
        t_assignment=lda[bow]
        tdists=[0 for i in range(NUM_TOPICS)]
        if DEBUG_LDA:
            print "Prepared text=",mtext
            print "BOW=", bow
            print "Assignment=", t_assignment
        
        for tass in t_assignment:
            tdists[tass[0]]=tass[1]
        
        for tdist in tdists:
            fl.write(","+str(tdist))
        
        fl.write(","+str(event["categories"]))
        fl.write("\n")
    
    fl.close()
   
    print "LDA assignment saved."
    print "done."


# Main code...
if __name__=="__main__":
    events=load_events(filename)
    stoplist=load_stoplist()
    generate_LDA_model(events, stoplist, filename, 100) #COMMENT here if you already have an topics file
                                                        #and want to use it
    generate_topic_assignment(events, stoplist, filename+"assignments.csv", "lda_25.model")
    